#dead or alive ultimate/online model exporter addon for blender version 0.4a
#all textures but the reflection one should have mesh uv layer
#reflection texture = "texture->affect->mirror"; transparent texture = "texture->affect->alpha"
#shadeless texture = "material->shading->shadeless";
#behind the hair(normals_mode = 8) = "material->use_tangent_shading"



bl_info = {
    "name": "Export doau xpr",
    "version": (1, 0),
    "blender": (2, 6, 3),
    "location": "File > Export > doau xpr",
    "description": "Export doau xpr",
    "category": "Import-Export"}

import bpy
from bpy.props import *
import mathutils, math, struct
from os import remove
import time
from bpy_extras.io_utils import ExportHelper
import subprocess
import inspect, os
import tempfile


#no spaces in the blender path or you're doomed
##tmpdir = tempfile.gettempdir()
ToolsPath = os.path.dirname(inspect.getfile(inspect.currentframe())) + "\\"
NvTriStripPath = ToolsPath + "NvTriStripper-cli.exe"
NvDxtPath = "nvdxt.exe"
cmdline = ' -nomipmap -outsamedir -rescale nearest -file temp.tga'
pwrs = [1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192]

def stripify(vert_indices):
	# Stripifying mesh
	NvTriStrip = subprocess.Popen([NvTriStripPath], stdin = subprocess.PIPE, stdout = subprocess.PIPE)

	for vi in vert_indices: NvTriStrip.stdin.write(str(vi).encode() + " ".encode())

	NvTriStrip.stdin.write("-1\n".encode())
	NvTriStrip.stdin.flush()

	stripcount = int(NvTriStrip.stdout.readline())
	if stripcount < 1: raise Exception("NvTriStrip returned 0 strips. Aborting")

	nativelist = []
	striptype = int(NvTriStrip.stdout.readline())
	nativelength = int(NvTriStrip.stdout.readline())
	nativelist.extend(map(int, NvTriStrip.stdout.readline().split()))
	return nativelist

class AnzVertex():
	x = y = z = nx = ny = nz = tu = tv = 0
	bones = []
	weights = []
	m = 0

	def __repr__(self):
		return self.__str__()

	def __str__(self):
		return "(%.3f %.3f %.3f | %.3f %.3f %.3f | %.3f %.3f)" %\
		(self.x, self.y, self.z, self.nx, self.ny, self.nz, self.tu, self.tv)

	def __eq__(self, v):
		return self.x == v.x and self.y == v.y and self.z == v.z \
		and self.nx == v.nx and self.ny == v.ny and self.nz == v.nz \
		and self.tu == v.tu and self.tv == v.tv \
		and self.bones == v.bones and self.weights == v.weights

	def __hash__(self):
		h = hash(self.x) ^ hash(self.y) ^ hash(self.z)\
		^ hash(self.nx) ^ hash(self.ny) ^ hash(self.nz)\
		^ hash(self.tu) ^ hash(self.tv)
		for b in self.bones:
			h ^= hash(b)
		for w in self.weights:
			h ^= hash(w)
		return h

class XPRnMDL():
    magicXPR0 = b'XPR0'
    xprsize = 0#vtxsize + indsize + 0x168 + allignto10 #dinamic
    block2ofs = 0#indsize + 0x168 + allignto10 #dinamic
    mdlid = 0x80000000
    mdlsize = 0#indsize + 0x168 - 0x24 #dinamic
    magicMDL = b'MDL\0'
    objnum = 0#1
    texnum = 0
    unknow = 0
    vtxbufnum = 0#1
    #objoffsets here
    objoffsetsend = 0
    objs = []
#mdl
#virtualdir
    vertbufid = 0x800001
    vertbufofs = 0
    texbufid = 0x40001
    bufpadding = 0
    listend = 0xFFFFFFFF
    #padding
    paddingtoblock2 = 0xAD
    #block2 (vertices)

    objects = []
    block2 = b''
    #mdlsize
    vdir = b''

    compiledxpr = b''
    objoffsets = []
    textures = []

    def __init__(self, textures_quality):
        self.xprsize = 0#vtxsize + indsize + 0x168 + allignto10 #dinamic
        self.block2ofs = 0#indsize + 0x168 + allignto10 #dinamic
        self.mdlsize = 0#indsize + 0x168 - 0x24 #dinamic
        self.objnum = 0#1
        self.texnum = 0
        self.unknow = 0
        self.vtxbufnum = 0#1
        self.objs = []
        self.vertbufofs = 0
        self.objects = []
        self.block2 = b''
        self.vdir = b''
        self.compiledxpr = b''
        self.objoffsets = []
        self.textures = []

        if bpy.context.selected_objects: bpy.ops.object.mode_set(mode='OBJECT')
        objnames = get_mesh_objects_names()
        for objname in objnames:
            newobj = OBJ(objname, objnames, self.textures)

            self.objects.append(newobj)
            self.objoffsets.append(self.mdlsize + 0x2C)#add the required offset
            self.mdlsize += newobj.objsize

            for i in range(4):
                if len(newobj.compiledvbufs[i]) > 0:
                    newobj.meshinfoslots[i]["vtxofs"] = len(self.block2)#add the offset to block2 later
                    self.block2 += newobj.compiledvbufs[i]
                    self.vdir += struct.pack("<LLL", self.vertbufid, newobj.meshinfoslots[i]["vtxofs"], self.bufpadding)
#handle images
        os.chdir(ToolsPath)
        for img, uses_alpha in self.textures:#compile textures
            bpy.context.scene.render.image_settings.file_format = 'TARGA'
            if uses_alpha:
                bpy.context.scene.render.image_settings.color_mode = 'RGBA'
                dxt = ' -dxt5'
                dxthex = 0xF
            else:
                bpy.context.scene.render.image_settings.color_mode = 'RGB'
                dxt = ' -dxt1c'
                dxthex = 0xC
            print(img)
            bpy.data.images[img].save_render('temp.tga')
            #print(cmdline + dxt)
            texquality = {0:' -quick', 1:' -quality_normal', 2:' -quality_production', 3:' -quality_highest'}[textures_quality]
            os.system(NvDxtPath+texquality + cmdline + dxt)
            os.remove('temp.tga')
            with open('temp.dds', 'rb') as f:
                sizeX, sizeY, nomipsize, magic = struct.unpack("<12x3L60x4s40x", f.read(128))
                mipmaps = 0
                flags = pwrs.index(sizeX)*0x1000000 | pwrs.index(sizeY)*0x100000 | mipmaps*0x10000 | dxthex*0x100 | 0x29
                imgofsinblock2 = len(self.block2)
                self.block2 += f.read(nomipsize)#image data to block2
                self.vdir += struct.pack("<5L", self.texbufid, imgofsinblock2, self.bufpadding, flags, self.bufpadding)
            os.remove('temp.dds')

        print("images:",len(self.textures))

        self.vdir += struct.pack("<L", self.listend)

        self.mdlsize = self.mdlsize + len(self.objects)*4 + 0x18

        self.block2ofs = self.mdlsize + 0x14 + len(self.vdir)
        allygnto10 = 0x100 - (self.block2ofs % 0x100)
        self.block2ofs += allygnto10
        for i in range(allygnto10):self.vdir += struct.pack("<B", self.paddingtoblock2)
        self.vtxbufnum = 0
        for objoffset in range(len(self.objoffsets)):
            self.objoffsets[objoffset] += len(self.objects)*4
            for i in range(4):
                if len(self.objects[objoffset].compiledvbufs[i]) > 0:
                    self.objects[objoffset].meshinfoslots[i]["idxofs"] += self.objoffsets[objoffset]
                    self.objects[objoffset].meshinfoslots[i]["vtxofs"] += self.block2ofs
                    self.vtxbufnum += 1
        self.xprsize = len(self.block2) + self.block2ofs
        self.objnum = len(self.objects)
        self.texnum = len(self.textures)

        self.compiledxpr += struct.pack("<4s4L4s4L", self.magicXPR0, self.xprsize, self.block2ofs, self.mdlid,
                                                    self.mdlsize, self.magicMDL, self.objnum, self.texnum,
                                                    self.unknow, self.vtxbufnum)
        for objoffset in self.objoffsets: self.compiledxpr += struct.pack("<L", objoffset)
        self.compiledxpr += struct.pack("<L", self.objoffsetsend)

        for obj in self.objects:
            self.compiledxpr += struct.pack("<4s3L4f", obj.magicOBJ, obj.objweight, obj.objunk, obj.objindxunused,
                                                    obj.bound[0], obj.bound[1], obj.bound[2], obj.bound[3])
            for i in range(4):
                self.compiledxpr += struct.pack("<4L", obj.meshinfoslots[i]["vtxnum"],
                                                        obj.meshinfoslots[i]["vtxofs"],
                                                        obj.meshinfoslots[i]["idxnum"],
                                                        obj.meshinfoslots[i]["idxofs"])
            self.compiledxpr += obj.compiledmats
        self.compiledxpr += self.vdir

class OBJ():
    magicOBJ = b'OBJ\0'
    objweight = 0
    objunk = 4
    objindxunused = 0 #dinamic
    bound = []#obj bounding sphere xc,yc,zc,rad
    meshinfoslots = [{"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},#four info slots
                    {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},
                    {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},
                    {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0}]

    # 0x40 of '0'
    matsend = 0
    magicINDEX = 0x12345678
    #all indicies go here
    vbuf = [[],[],[],[]]
    ibuf = [[],[],[],[]]
    compiledmats = b''
    compiledvbufs = [b'',b'',b'',b'']
    objsize = 0
    def __init__(self, objname, objnames, textures):
        self.bound = []#obj bounding sphere xc,yc,zc,rad
        self.meshinfoslots = [{"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},
                        {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},
                        {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0},
                        {"vtxnum":0, "vtxofs":0, "idxnum":0, "idxofs":0}]
        self.vbuf = [[],[],[],[]]
        self.ibuf = [[],[],[],[]]
        self.compiledmats = b''
        self.compiledvbufs = [b'',b'',b'',b'']
        self.objsize = 0


        #compille buffers padding + mats
        for i in range(0x10): self.compiledmats += struct.pack("<L", 0)

        print(objname)
        bpy.ops.object.select_all(action='DESELECT')
        bpy.data.objects[objname].select = True
        bpy.context.scene.objects.active = bpy.data.objects[objname]

        bpy.ops.object.duplicate()#the new copy is selected
        bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)#apply changes

        #create bounding sphere
        x0,y0,z0 = bpy.context.selected_objects[0].bound_box[0]
        x1,y1,z1 = bpy.context.selected_objects[0].bound_box[6]
        rad =  math.sqrt((x0-x1)*(x0-x1) + (y0-y1)*(y0-y1) + (z0-z1)*(z0-z1))/2 # inacurate
        rad *= 1.1
        xc,yc,zc = (x1+x0)/2, (y1+y0)/2, (z1+z0)/2
        self.bound = [xc, zc, -yc, rad] #rotate axis

        bpy.ops.object.editmode_toggle()
        bpy.ops.mesh.separate(type='MATERIAL')#if no material it's cancelled
        bpy.ops.object.editmode_toggle()

        tempobjnames = get_mesh_objects_names()
        matobjnames = [item for item in tempobjnames if item not in objnames]#materials objects names
        #print("separated in",len(matobjnames))
        for matobjname in matobjnames:
##            if bpy.data.objects[matobjname].data.materials and 'NULL' in bpy.data.objects[matobjname].data.materials[ bpy.data.objects[matobjname].data.polygons[0].material_index ].name.upper():
##                continue #if 'null' text is found in the name of the material then this material is skipped (use this kind of materials to create correct normals on the edges of the smooth groups)
            self.compiledmats += processMAT(matobjname, self.vbuf, self.ibuf, textures)
        self.compiledmats += struct.pack("<LL", self.matsend, self.magicINDEX)
        for i in range(4):
            if len(self.vbuf[i]) > 0:
                self.meshinfoslots[i]["idxofs"] = len(self.compiledmats) + 0x60
                for H in self.ibuf[i]:
                    self.compiledmats += struct.pack("<H", H)

                self.meshinfoslots[i]["vtxnum"] = len(self.vbuf[i])
                self.meshinfoslots[i]["idxnum"] = len(self.ibuf[i])
                if len(self.vbuf[i]) > 0xFFFF: print("ERROR: an object can hold only up to 65535 vertices. currently", self.vtxnum)

                if i == 0:
                    for v in self.vbuf[i]:
                        self.compiledvbufs[i] += struct.pack("<8f", v.x, v.z, -v.y, #rotate axis orientation
                                                                v.nx, v.nz, -v.ny, #normales axis fixed?
                                                                v.tu, v.tv)
                elif i == 1:
                    for v in self.vbuf[i]:
                        self.compiledvbufs[i] += struct.pack("<6f", v.x, v.z, -v.y, #rotate axis orientation
                                                                1.0, #diffuse power should be taken from blender?
                                                                v.tu, v.tv)
        self.objsize = len(self.compiledmats) + 0x60


def processMAT(objname, vbuf, ibuf, textures):
#TODO - if shadeless then: second "mesh slot", vertsise and bufnvsize = crresponding to slot,
    activemeshinfoslot = 0

    matid = 0x80000000# 0x80000001 = transparrent
    matsize = 0x90 #(0x80 + one texture size)dinamic in perspective(when will handle textures)
    matunk0 = 0
    nrm_mode = 0 #normales direction: 0 - both sides, 1 - normal, 2 - inverse, 8 - both sides(+behind the characters hair), others - ?
##    bound = []#mat bounding sphere

    diffuse = []
    ambient = []
    specular = []
    emissive = []
    specular_glow_power = 50.0
    matimagesnum = 1#one texture
    matlight = 1
    matff = 0xFFFFFFFF


    bufnvsize = (0x00020, 0x10018, 0x20028, 0x30020)
    indstart = 0 #dinamic
    indend = 0#dinamic
    indunused = 2

    #start processing

    bpy.ops.object.select_all(action='DESELECT')
    bpy.data.objects[objname].select = True
    bpy.context.scene.objects.active = bpy.data.objects[objname]
    bpy.ops.object.editmode_toggle()
    bpy.ops.mesh.select_all(action = 'SELECT')
    bpy.ops.mesh.quads_convert_to_tris()
    bpy.ops.object.editmode_toggle()

    #create bounding sphere(inacurate radius coz it's based on the bounding box)
    x0,y0,z0 = bpy.data.objects[objname].bound_box[0]
    x1,y1,z1 = bpy.data.objects[objname].bound_box[6]
    rad =  math.sqrt((x0-x1)*(x0-x1) + (y0-y1)*(y0-y1) + (z0-z1)*(z0-z1))/2 * 1.1
    xc,yc,zc = (x1+x0)/2, (y1+y0)/2, (z1+z0)/2
##    bound = [xc, yc, zc, rad]

    mesh = bpy.data.objects[objname].data
    mesh.update(calc_tessface = True)#initialize tessfaces
    has_uv = True if mesh.tessface_uv_textures.active else False

    diffuse = [1.0, 1.0, 1.0, 1.0]
    ambient = [1.0, 1.0, 1.0, 1.0]
    specular = [1.0, 1.0, 1.0, 1.0]
    emissive = [0.0, 0.0, 0.0, 0.0]
    specular_glow_power = 50.0
    diffuse_power = 1.0#used for doao shadeless textures
    transparency = False
    matimages = []
    if bpy.data.objects[objname].material_slots:
        material = mesh.materials[bpy.data.objects[objname].data.polygons[0].material_index]
        if material.use_shadeless:
            activemeshinfoslot = 1#x,y,z,diffuse,u,v
        if material.use_tangent_shading:
            nrm_mode = 8#behind the hair
        btexslots = material.texture_slots
##        for i in range(1): #range(3): #check 3 texture slots in blender material
        if btexslots[0] and btexslots[0].texture.type == "IMAGE":
            texid = 0x888# x888 = null texture
            texun1 = 0
            textype = 0#1 = transparent
            texflags = 0x400
            #if has_uv:
            if btexslots[0].use_map_alpha:#textures with transparency are recognized by this
                transparency = True
                print("+", end = " ")
                textype = 1#transparent rexture
                matid = 0x80000001 #transparent material
            elif btexslots[0].use_map_mirror:#mirror texture(not for flat mehes), do not require uv layer
                texflags = 0x1000400
            img = btexslots[0].texture.image.name
            print(img, "" if has_uv else "no uv layer found")
            if len(img) > 4 and ".000" < img[-4:] < ".999":#only one image for all imported instances
                img = img[:-4]
                print(img)
            if img not in [i[0] for i in textures]:#get only names from names/transparency
                textures.append([img, transparency])
            texid = [i[0] for i in textures].index(img)
            if transparency == True and textures[texid][1] == False:
                    textures[texid][1] = True#if at least once is transparent then it's transparent

            if not (btexslots[1] and btexslots[1].texture.type == "IMAGE" and btexslots[1].use_map_mirror and not btexslots[0].use_map_mirror):#if env map (should be semitransparrent)
                matimages.append((texid, texun1, textype, texflags))
            else:
                matimages.append((texid, texun1, 3, texflags))
                img = btexslots[1].texture.image.name
                print(img, "" if has_uv else "no uv layer found")
                if len(img) > 4 and ".000" < img[-4:] < ".999":#only one image for all imported instances
                    img = img[:-4]
                    print(img)
                if img not in [i[0] for i in textures]:#get only names from names/transparency
                    textures.append([img, transparency])
                texid = [i[0] for i in textures].index(img)
                textures[texid][1] = True#this env map texture should be transparent to show the diffuse texture
                matimages.append((texid, texun1, 4, 0x1000400))
                matimages.append((0x888, texun1, 5, 0))

        if not matimages:
            matimages.append((0x888, 0, 0, 0x400))


            diffuse[0:3] = material.diffuse_color
            #diffuse[3] = material.alpha #don't use this it can give you trouble
            ambient[0:3] = material.mirror_color
            ambient[3] = material.ambient
            emissive[3] = material.emit
            #diffuse_power = material.translucency
            print(material.name, end = " ")
        specular[0:3] = material.specular_color
        specular_glow_power = material.specular_intensity

    worldMatrix = bpy.data.objects[objname].matrix_world
    # Making native faces
    vertMap = {}
    vertUniqueList = []
    vertIndexList = []
    vCount = 0
    for face in mesh.polygons:
    	vi = 0
    	if(len(face.vertices) != 3): print("not a triangle ", (face.vertices[:]))
    	for i in range(len(face.vertices)):
    		v = AnzVertex()
    		wV = mesh.vertices[face.vertices[i]].co * worldMatrix
    		v.x = wV.x
    		v.y = wV.y
    		v.z = wV.z
    		if face.use_smooth:
    		    v.nx, v.ny, v.nz = mesh.vertices[face.vertices[i]].normal
    		elif activemeshinfoslot == 1:
        		v.nx = v.ny = v.nz = 0.0
    		else:
        		v.nx, v.ny, v.nz = face.normal

    		if has_uv:
    		    v.tu = mesh.tessface_uv_textures.active.data[face.index].uv[i][0]
    		    v.tv = 1 - mesh.tessface_uv_textures.active.data[face.index].uv[i][1]
    		else:
        		v.tu = v.tv = 0

    		if v in vertMap:
    			vertUniqueList.append(vertMap[v])
    		else:
    			newIndex = vCount
    			vertUniqueList.append(newIndex)
    			vertMap[v] = newIndex
    			vertIndexList.append(v)
    			vCount += 1
    		vi += 1

    # Stripifying mesh
    print("........  verts =", len(vertIndexList), " indicies =", len(vertUniqueList), end = "")
    if len(vertIndexList) > 65535: print("\nERROR: nvtristrip can handle only up to 65535 vertices")
    nativelist = stripify(vertUniqueList)
    print(" =>", len(nativelist))


    indstart = len(ibuf[activemeshinfoslot])
    indend = len(nativelist) - 2

    templist = [i+len(vbuf[activemeshinfoslot]) for i in nativelist]
    ibuf[activemeshinfoslot].extend(templist)
    vbuf[activemeshinfoslot].extend(vertIndexList)

    #handle texture

    bpy.ops.object.delete()
    bpy.data.meshes.remove(mesh)

    matimagesnum = len(matimages)#number of used textures
    matsize = 0x80 + matimagesnum*0x10
    #compile material
    result = struct.pack("<4L4f17f3L",  matid,  matsize,  matunk0,  nrm_mode,#head
                            xc, zc, -yc, rad,#bound #rotate axis
                             diffuse[0],  diffuse[1],  diffuse[2],  diffuse[3],#colors
                             ambient[0],  ambient[1],  ambient[2],  ambient[3],
                             specular[0],  specular[1],  specular[2],  specular[3],
                             emissive[0],  emissive[1],  emissive[2],  emissive[3],
                             specular_glow_power,  matimagesnum,  matlight,  matff)#textures
    for texid, texun1, textype, texflags in matimages:
        result += struct.pack("<4L",  texid, texun1, textype, texflags) #one texture added

    result += struct.pack("<4L",  bufnvsize[activemeshinfoslot],  indstart,  indend,  indunused)#end
    return result

def get_mesh_objects_names():
#    bpy.ops.object.select_by_type(type='MESH')
#    objnames = [obj.name for obj in bpy.context.selected_objects]
#    if not objnames: raise Exception("Nothing to export.")
    objnames = [obj.name for obj in bpy.data.objects if obj.type == 'MESH']
    if not objnames: raise Exception("Nothing to export.")
    return objnames

def do_export(context, props, filepath, textures_quality):
    print(NvTriStripPath)
##    if os.path.isfile(ToolsPath + NvDxtPath):
##        print(ToolsPath + NvDxtPath)
##    else:
##        print(ToolsPath + NvDxtPath)
##        print("NvDxt is not found on ", 1/0)
    myxpr = XPRnMDL(textures_quality)
    with open(filepath, "wb") as w:
        w.write(myxpr.compiledxpr)
        w.write(myxpr.block2)
    #del myxpr
    myxpr = None
    return True


###### EXPORT OPERATOR #######
class Export_pc2(bpy.types.Operator, ExportHelper):
    '''Exports the active Object as a doau xpr file'''
    bl_idname = "export_doau.xpr"
    bl_label = "Export doau xpr"
    filename_ext = ".xpr"

    textures_quality = EnumProperty(
            name="Textures Quality",
            items=(("3", "highest", "highest quality compression (this can be very slow)"),
                   ("2", "production", "production quality compression"),
                   ("1", "normal", "normal quality compression"),
                   ("0", "quick", "use fast compression method"),
                   ),
            default='1',
            description = "Quality of exported textures",
            )

    def execute(self, context):
        start_time = time.time()
        print('\n_____START_____')
        props = self.properties
        filepath = self.filepath
        filepath = bpy.path.ensure_ext(filepath, self.filename_ext)

        exported = do_export(context, props, filepath, int(self.textures_quality))

        if exported:
            print('finished export in %s seconds' %((time.time() - start_time)))
            print(filepath)

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager

        if True:
            # File selector
            wm.fileselect_add(self) # will run self.execute()
            return {'RUNNING_MODAL'}
        elif True:
            # search the enum
            wm.invoke_search_popup(self)
            return {'RUNNING_MODAL'}
        elif False:
            # Redo popup
            return wm.invoke_props_popup(self, event)
        elif False:
            return self.execute(context)


### REGISTER ###
def menu_func(self, context):
    self.layout.operator(Export_pc2.bl_idname, text="doau xpr")

def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_export.append(menu_func)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_export.remove(menu_func)

if __name__ == "__main__":
    register()
